module OW_3drot
! 3D-linear Radon transform

  use OW_type
  use OW_sltpar

  integer, private		:: oxchunk,xblk
  real, allocatable, private  :: dips(:,:)

  contains

!===============================================================================
  function ow3drot_init(oxchunk_in,xblk_in,dips_in) result(stat)
    integer                     :: oxchunk_in,xblk_in,stat
    real,optional               :: dips_in(:,:)
    oxchunk=oxchunk_in
    xblk=xblk_in
    if (present(dips_in)) then
       allocate(dips(nz,xblk))
       dips=dips_in
    end if
    stat=0
  end function 
!===============================================================================
  function ow3drot(adj, add, modl, data) result(stat)
    integer                     :: stat 
    logical,intent(in)          :: adj,add 
    real,dimension(:)           :: modl
    real,dimension(:)           :: data

    if (adj) then
       modl=0.
    else
       data=0.
    end if
    call ow3drot_exe(adj,add,modl,data )
    stat=0
  end function 
!=================================================================================  
  subroutine ow3drot_exe(adj,add,modl,data)
    logical,intent(in)        :: adj,add
    real		  	:: modl(:) 
    real		  	:: data(:)

!! internal variables
    integer		:: inode
    integer 		:: iax,ix,ixm,iah,ih,izm,izd,idx
    integer		:: minx,maxx,indd,indm,idhx
    real			:: ax,tgx,xx,ah,tgh,hx,zm,zd
    real                      :: hw,tgs,rdx,fx,gx,dhx,dhz,y

    integer, external         :: omp_get_thread_num,omp_get_num_threads

if (nodes.eq.9999) then
!$OMP PARALLEL
    nodes = omp_get_num_threads()
!$OMP END PARALLEL
end if

!$OMP PARALLEL DO PRIVATE  (inode,&
       hw,y,iax,ix,iah,ih,izm,izd,&
      idx,indd,indm,tgs,rdx,fx,gx,&
dhx,idhx,dhz,ax,tgx,ah,tgh,hx,zm,zd) schedule(dynamic,1)  num_threads(nodes) 
                               
do ix=1,xblk+1
   inode = omp_get_thread_num()+1
   do iax=1,nax
      ax = oax + (iax-1)*dax
      tgx = -tand(ax)
      do iah=1,nah
         ah = oah + (iah-1)*dah
         tgh = -tand(ah)
         if (rec) then
            if (1-tgx*tgh.le.tiny(1.)) then 
               cycle
            else
               tgs=(tgx+tgh)/(1-tgx*tgh)
            end if
         else
            if (1+tgx*tgh.le.tiny(1.)) then
               cycle
            else
               tgs=(tgx-tgh)/(1+tgx*tgh)
            end if
         end if
         do ih=1,sog%sh%n
            hx=(ih-1)*sog%sh%d+sog%sh%o
            hw=1 - abs(hx)/2/abs(sog%sh%o)
            hw=hw**4
!!
!            dhx=abs(hx)*cosd(ax)*tgh*tgx
!            idhx = 1.5 + dhx/sog%sh%d
!            dhz=abs(hx)*sind(ax)*tgh*tgx
            dhx=abs(hx)*sind(ax)*cosd(ax)*abs(tgh)
            idhx = 1.5 + dhx/sog%sh%d
            dhz=abs(hx)*sind(ax)*sind(ax)*abs(tgh)
!            dhx=abs(hx)*sind(ax)*cosd(ax)*tgh
!            idhx = 1.5 + dhx/sog%sh%d
!            dhz=abs(hx)*sind(ax)*sind(ax)*tgh
!!
            do izm=1,nz
               if (allocated(dips)) then
                  if (ix.eq.xblk+1) then
                     y = (1 - abs(dips(izm,xblk)-ax)/2/abs(oax))
                  else
                     y = (1 - abs(dips(izm,ix)-ax)/2/abs(oax))
                  end if
                  if (y < 0) y=0.
                  y = y**6
               else
                  y = 1           
               end if
               zm  = zmin + (izm-1) * sog%z%d
               if (rec) then
                  if (upshft) then
                     zd  = zm + tgs*hx + dhz
                  else
                     zd  = zm + tgs*hx
                  end if
               else
                  if (upshft) then
                     zd  = zm - tgs*hx + dhz
                  else
                     zd  = zm - tgs*hx
                  end if
               end if
!!.. linear interpolation
!               rdx = (zd - sog%z%o) / sog%z%d
!               idx = rdx
!               izd = idx + 1
!               if (izd > sog%z%n.or. izd < 1) cycle
!               fx = rdx - idx
!               gx = 1.0 - fx
!               if (indm > size(modl) .or. indm < 1 .or. indd+1 > size(data) .or. indd+1 < 1) cycle
!               if (adj) then
!                  modl(indm) = modl(indm) +  gx*data(indd) +fx*data(indd+1)
!               else
!                  data(indd+1)  = data(indd+1)  + modl(indm)*fx
!                  data(indd  )  = data(indd  )  + modl(indm)*gx
!               end if 
!....................
!               izd = 1.5 + (zd-sog%z%o)/sog%z%d
!               if ( izd > sog%z%n .or. izd < 1) cycle
               izd = 1.5 + (zd-zmin)/sog%z%d
               if ( izd > nz .or. izd < 1) cycle
               indm=(ix-1)*nz*sog%sh%n+(ih-1)*nz+izm
!               if (indm > (xblk-1)*nz*sog%sh%n+(ih-1)*nz+izm .and. &
!                   indm > (xblk-1)*nz*sog%sh%n+(ih-1)*nz+nz) write(0,*) indm
               if (upshft) then
                  indd=(ix+idhx-1)*nz*sog%sh%n*nax*nah+(iax-1)*nz*sog%sh%n*nah+(iah-1)*nz*sog%sh%n+(ih-1)*nz+izd
               else
                  indd=(ix-1)*nz*nax*nah+(iax-1)*nz*nah+(iah-1)*nz+izd
               end if
               if (indm > size(modl) .or. indm < 1 .or. indd > size(data) .or. indd < 1) cycle
               if (adj) then
                  modl(indm) = modl(indm) +  y*data(indd)/sqrt(abs(cos(atan(tgs))))/sqrt(hw)
               else
                  data(indd) = data(indd) + y*modl(indm)/sqrt(abs(cos(atan(tgs))))/sqrt(hw)
               end if 
            end do
         end do
      end do
   end do
end do

!$OMP END PARALLEL DO
 write(0,*) "DONE ROTATION UP TO CIG", oxchunk-1+xblk, " FROM ", sog%xm%n
 end subroutine
!===============================================================================
  function ow3drot_close() result(stat)
    deallocate(dips)
    stat=0
  end function 
!=================================================================================  

end module 
