module OW_wei_arealoldok

use sep
use OW_type
use OW_image_areal
use OW_fftw_areal
use OW_parms_areal
use OW_pspi_areal
use OW_ffd_areal
use OW_scttr_areal
!use OW_cstret
use triang
use sep3d_struct_mod

implicit none

integer, private                             :: idip,nodes
type(sep3d), private                         :: aximg

contains
!======================================================================
subroutine owwei_init(idip_in)
integer,optional                        :: idip_in
integer, external			:: omp_get_thread_num,omp_get_num_threads
if (node.eq.9999) then
  !$OMP PARALLEL
  nodes = omp_get_num_threads()
  !$OMP END PARALLEL
else
  nodes=node
end if
write(3,*) "****************"
write(3,*) "* NUM.THREADS = ",nodes
write(3,*) "****************"
if(present(idip_in)) then
   idip=idip_in
else
   idip=1
end if
end subroutine
!======================================================================
subroutine owwei(taper,recv,source,vel,image,illum,sref)

real, optional					:: sref(:,:,:)
complex						:: recv(:,:),source(:,:)
integer						:: iwa,iz,inode,i,ixw,ish
real						:: taper(:),vel(kxm%n,img%z%n),a
real						:: image(:,:,:),illum(kxm%n,img%z%n),illumina(kxm%n,img%z%n,nodes)
real						:: GGG(kxm%n,nsecdim,nz,nodes)
complex						:: ws(kxm%n,nodes),wr(kxm%n,nodes)
complex						:: tflds(kxm%n,nodes),tfldr(kxm%n,nodes)
integer, external					:: omp_get_thread_num,omp_get_num_threads

wr=0
ws=0
GGG=0

!--------------------------------------------------
! Setting FFTW plans
tflds=0
call owfftws(tflds,nodes)
tfldr=0
call owfftwr(tfldr,nodes)
!$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
   do iwa = 1,rec%w%n
      inode=omp_get_thread_num()+1
      ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=source(:,iwa)
      wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=recv(:,iwa)
      do iz=max(2,nzmin),nzmax
         if (extrap(1:3).eq."ffd") then
            if (rec%w%all(iwa).ne.0) then
               call owffd( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
               call owffd( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
            end if
         else if (extrap(1:4).eq."pspi") then
            if (present(sref)) then
               call owpspi_init(iz)
               call owpspi( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
               call owpspi( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
            else
               call owpspi_init(iz)
               call owpspi( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
               call owpspi( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
            end if
         end if
         ws(:,inode)=ws(:,inode)*taper
         wr(:,inode)=wr(:,inode)*taper
         illumina(:,iz,inode)=illumina(:,iz,inode)+ws(:,inode)*conjg(ws(:,inode))
         call owimage(.false., iwa, ws(:,inode),  wr(:,inode), GGG(:,:,iz-nzmin+1,inode), vel(:,iz-1))
      end do
      if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
				" FROM NODE: ",inode
   end do
!$OMP END PARALLEL DO
   do i=1,nodes
      !$OMP parallel do private(ixw,iz,ish)
      do iz=1,nz
         illum(:,iz)=illum(:,iz)+illumina(:,iz,i)
         do ish=1,nsecdim
            do ixw=1,kxm%n
               image(ixw,ish,iz)=image(ixw,ish,iz)+GGG(ixw,ish,iz,i)
            end do
         end do
      end do
      !$OMP END PARALLEL DO
   end do
6 format(A13,i3,A8,i3,A14,i3,A12,i3)
call owfftdestroy()
end subroutine
!======================================================================
subroutine owzwht(wgcig)
real                                                        :: wgcig(:,:,:)
integer                                                     :: iz,ix,ish
real		                                        :: temp(img%z%n)

   temp=0.
   do ish=1,nsecdim;do ix=1,img%xm%n
      temp=abs(wgcig(ix,ish,:))
      do iz=5,img%z%n-4
         temp(iz)=sum(temp(iz-4:iz+4))/9
      end do
      wgcig(ix,ish,:)=temp*wgcig(ix,ish,:)
   end do;end do
end subroutine
!======================================================================
subroutine owzlinwht(wgcig)

real                                                        :: wgcig(:,:,:)
integer                                                     :: iz

do iz=1,img%z%n
   wgcig(:,:,iz)=wgcig(:,:,iz)*float(iz)**zwgt
end do
end subroutine
!======================================================================
subroutine owsmooth(wgcig,odcig)
real                                                        :: wgcig(:,:,:),odcig(:,:,:)
integer                                                     :: iz,ix,ish,iwi,iwf
real,allocatable	                                        :: temp(:)

allocate(temp(img%z%n))
   temp=0.;odcig=0.
   do ish=1,nsecdim;do ix=1,img%xm%n;
      do iz=1,img%z%n
         iwi=max(iz-windz+1,1)
         iwf=min(iz+windz-1,img%z%n)
         temp(iz)=sum(abs(wgcig(ix,ish,iwi:iwf)))/(iwf-iwi+1)
      end do
      odcig(ix,ish,:)=temp
   end do;end do

deallocate(temp); allocate(temp(img%xm%n))
   temp=0.
   do iz=1,img%z%n;do ish=1,nsecdim;
      do ix=1,img%xm%n
         iwi=max(ix-windx+1,1)
         iwf=min(ix+windx-1,img%xm%n)
         temp(ix)=sum(odcig(iwi:iwf,ish,iz))/(iwf-iwi+1)
      end do
      odcig(:,ish,iz)=temp
   end do;end do

deallocate(temp); allocate(temp(img%sh%n))
   temp=0.
   do iz=1,img%z%n;do ix=1,img%xm%n
      do ish=1,nsecdim
         iwi=max(ish-windh+1,1)
         iwf=min(ish+windh-1,img%sh%n)
         temp(ish)=sum(odcig(ix,iwi:iwf,iz))/(iwf-iwi+1)
      end do
      odcig(ix,:,iz)=temp
   end do;end do
deallocate(temp)
end subroutine
!======================================================================
subroutine owxwht(adj,wgcig,ofcig)
logical                                                     :: adj
real                                                        :: wgcig(:,:,:),ofcig(:,:,:)
integer                                                     :: iz,ix,ish
real		                                        :: temp(img%xm%n)

if (adj) then
   temp=0.;wgcig=0.
   do iz=1,img%z%n;do ish=1,nsecdim
      temp=ofcig(:,ish,iz)
      do ix=int(float(nxfilt)/2.)+1,img%xm%n-int(float(nxfilt)/2.)
         temp(ix)=sum(temp(ix-int(float(nxfilt)/2.):ix+int(float(nxfilt)/2.)))/nxfilt
      end do
      wgcig(:,ish,iz)=wgcig(:,ish,iz)+temp
   end do;end do
else
   temp=0.;ofcig=0.
   do iz=1,img%z%n;do ish=1,nsecdim
      temp=wgcig(:,ish,iz)
      do ix=int(float(nxfilt)/2.)+1,img%xm%n-int(float(nxfilt)/2.)
         temp(ix)=sum(temp(ix-int(float(nxfilt)/2.):ix+int(float(nxfilt)/2.)))/nxfilt
      end do
      ofcig(:,ish,iz)=ofcig(:,ish,iz)+temp
   end do;end do
end if
end subroutine
!======================================================================
subroutine owdso(adj,wgcig)!,ofcig)
logical						:: adj
real						:: wgcig(:,:,:)!,ofcig(:,:,:)
integer						:: iz,ix,ish
real						:: h(nsecdim)

   do ish=1,nsecdim
      h(ish)=(ish-1)*dsecdim+osecdim
   end do
   h = (abs(h / maxval(h)))**expdso

   do ish=1,nsecdim;do ix=1,img%xm%n
      wgcig(ix,ish,:)=wgcig(ix,ish,:)*h(ish)!*ofcig(ix,ish,:)
   end do;end do
end subroutine
!======================================================================
subroutine owenvelope( B_image)

  integer                   :: i1,i2,i3
  real                      :: B_image(:,:,:)
  real                      :: bot(size(B_image,2),img%z%n), data(img%xm%n,size(B_image,2),img%z%n),a,b
  complex                   :: cdata(img%z%n,1)

  cdata=0.
  call owfftwr(cdata,1)
  do i2= 1, img%xm%n  
    do i3=1,size(B_image,2)
      cdata(:,1) = cmplx(B_image(i2,i3,:),0.)
      call owfftexer(cdata(:,1),1)
      cdata(:,1)=cdata(:,1)/sqrt(1.*img%z%n)
      cdata(1,1)      = cdata(1,1)      / 2.
      cdata(1+img%z%n/2,1) = cdata(1+img%z%n/2,1) / 2.
      cdata(img%z%n/2+2:img%z%n,1) = 0.
      call owifftexer(cdata(:,1),1)
      cdata(:,1)=cdata(:,1)/sqrt(1.*img%z%n)
      bot(i3,:) = conjg( cdata(:,1)) * cdata(:,1)
    end do 
    call triangle3( tc2, tc1, nsecdim, img%z%n, bot, bot)
    do i3=1,size(B_image,2)
      where(bot(i3,:).eq.0) bot(i3,:)=1.
    end do
    data(i2,:,:) = sqrt( bot)
  end do 
  a=0
  do i2= 1, img%xm%n
     do i3=1,nsecdim  
        if (a < maxval(data(i2,i3,:))) then
           a = maxval(data(i2,i3,:))
        end if  
     end do
  end do
  B_image=(data+a)**2
!  B_image=B_image/data
!  a=0;b=0
!  do i2= 1, img%xm%n  
!    do i3=1,nsecdim  
!       a=maxval(data(i2,i3,:))
!       b=minval(data(i2,i3,:))
!       write(4,*) i2,i1,a,b
!       a=maxval(data(i2,i3,:))
!       data(i2,i3,:)=data(i2,i3,:)/a
!       where(data(i2,i3,:).ge.0.5) data(i2,i3,:)=1
!       where(data(i2,i3,:).le.0.25) data(i2,i3,:)=0
!    end do
!  end do
!  B_image=B_image*data
end subroutine 
!======================================================================
subroutine owzphase(wgcig)

real						:: wgcig(:,:,:)
complex						:: phase(img%z%n),tfld(img%z%n)
integer						:: iz,ix,ish

phase = exp(cmplx(cos(pi/4.),sin(pi/4.))*sqrt(abs(kz%all)))
write(0,*) phase
call owfftw(tfld)
do ish=1,nsecdim
   do ix=1,img%xm%n
      tfld=cmplx(wgcig(ix,ish,:),0)
      call sfftw_execute(planF)
      tfld=tfld/sqrt(1.*img%z%n)*phase
      call sfftw_execute(planI)
      wgcig(ix,ish,:)=-real(tfld/sqrt(1.*img%z%n))
   end do
end do
call sfftw_destroy_plan(planF)
call sfftw_destroy_plan(planI)
end subroutine
!======================================================================
subroutine owwemva_init(idip_in)
integer      :: idip_in
integer, external			:: omp_get_thread_num,omp_get_num_threads
if (node.eq.9999) then
  !$OMP PARALLEL
  nodes = omp_get_num_threads()
  !$OMP END PARALLEL
else
  nodes=node
end if
write(3,*) "****************"
write(3,*) "* NUM.THREADS = ",nodes
write(3,*) "****************"
idip=idip_in
end subroutine
!======================================================================
subroutine owwemva(adj,taper,recv,source,vel,deltaS,deltaI,sref)

real, optional					:: sref(:,:,:)
logical						:: adj
complex						:: recv(:,:),source(:,:)
integer						:: iwa,iz,inode,i,ixw,ish,n0
real						:: taper(:),vel(kxm%n,img%z%n)
real						:: deltaS(kxm%n,img%z%n)
real						:: deltaI(kxm%n,nsecdim,img%z%n)
real						:: GGG(kxm%n,img%z%n,nodes)
real						:: GG(kxm%n,nsecdim,img%z%n,nodes)
complex						:: ws(kxm%n,nodes),wr(kxm%n,nodes)
complex						:: ddws(kxm%n,nodes),ddwr(kxm%n,nodes),temp(kxm%n,nodes)
complex, allocatable				:: dws(:,:,:),dwr(:,:,:),DDD(:,:,:)
complex						:: bws(kxm%n,img%z%n,nodes),bwr(kxm%n,img%z%n,nodes)
complex						:: tflds(kxm%n,nodes),tfldr(kxm%n,nodes)
complex						:: dws_z(kxm%n,nodes),dwr_z(kxm%n,nodes)
complex 						:: zws(kxm%n,img%z%n,nodes),zwr(kxm%n,img%z%n,nodes)
integer, external					:: omp_get_thread_num,omp_get_num_threads

wr=0
ws=0
ddwr=0
ddws=0
bwr=0
bws=0

!--------------------------------------------------
! Setting FFTW plans
tflds=0
call owfftws(tflds,nodes)
tfldr=0
call owfftwr(tfldr,nodes)

if (adj) then
   GGG=0
   allocate(dws(kxm%n,img%z%n,nodes),dwr(kxm%n,img%z%n,nodes))
   dwr_z=0
   dws_z=0
   dwr=0
   dws=0
  
   !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
   do iwa = 1,rec%w%n
       inode=omp_get_thread_num()+1
       ws(:,inode)=0.;wr(:,inode)=0.
       ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=source(:,iwa)
       wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=recv(:,iwa)
       dws(:,:,inode)=0.
       dwr(:,:,inode)=0.
       bws(:,:,inode)=0.
       bwr(:,:,inode)=0.
       dws_z(:,inode)=0.
       dwr_z(:,inode)=0.
       ddwr(:,inode)=0
       ddws(:,inode)=0
       bws(:,1,inode)=ws(:,inode)
       bwr(:,1,inode)=wr(:,inode)

!!!!!!!!!!!!!!!!!!!!!!!
!
! Background wavefields
! bws and bwr
!
      do iz=max(2,nzmin),nzmax
         if (extrap(1:3).eq."ffd") then
            if (rec%w%all(iwa).ne.0) then
               call owffd( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
               call owffd( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
            end if
         else if (extrap(1:4).eq."pspi") then
            if (present(sref)) then
               call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
               call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
            else
               call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
               call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
            end if
         end if
          ws(:,inode)=ws(:,inode)*taper
          wr(:,inode)=wr(:,inode)*taper
          bws(:,iz,inode)=ws(:,inode)
          bwr(:,iz,inode)=wr(:,inode)
       end do


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Back project image perturbation
       do iz=img%z%n,max(2,nzmin),-1
          dws=0.
          call owimage(.true.,  iwa, dws(:,iz,inode),  bwr(:,iz,inode), deltaI(:,:,iz), vel(:,iz-1), .true. )
          dws_z(:,inode)=dws_z(:,inode)+dws(:,iz,inode)
          if (born2) then
             call owimage(.true., iwa, dws(:,iz,inode), temp(:,inode), deltaI(:,:,iz),vel(:,iz-1), .false.)
             dwr_z(:,inode)=dwr_z(:,inode)+temp(:,inode)
             temp(:,inode)=0
          end if
          dws_z(:,inode)=dws_z(:,inode)*taper
          if (extrap(1:3).eq."ffd") then
             if (rec%w%all(iwa)*10.ne.0) then
                call owffd( iwa, inode, -1,  dws_z(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
             end if
          else if (extrap(1:4).eq."pspi") then
             if(present(sref)) then
                call owpspi(  iwa, inode, -1, dws_z(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
             else
                call owpspi(  iwa, inode, -1, dws_z(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
             end if
          end if
          if (renorm) then
             bws(:,iz-1,inode)=bws(:,iz-1,inode)+dws_z(:,inode)
!             bws(:,iz,inode)=bws(:,iz,inode)+dws_z(:,inode)
          end if
          ddws(:,inode)=0
          if (rec%w%all(iwa).ne.0.) then
             call owscctrx(iwa, 1,ddws(:,inode),bws(:,iz-1,inode),vel(:,iz-1))
          end if
          GGG(:,iz,inode)=GGG(:,iz,inode)+conjg(ddws(:,inode))*dws_z(:,inode)

          call owimage(.true.,  iwa, bws(:,iz,inode),  dwr(:,iz,inode), deltaI(:,:,iz), vel(:,iz-1), .false.)
          dwr_z(:,inode)=dwr_z(:,inode)+dwr(:,iz,inode)
          if (born2) then
             call owimage(.true., iwa, temp(:,inode), dwr(:,iz,inode), deltaI(:,:,iz),vel(:,iz-1), .true.)
             dws_z(:,inode)=dws_z(:,inode)+temp(:,inode)
             temp(:,inode)=0
          end if
          dwr_z(:,inode)=dwr_z(:,inode)*taper
          if (extrap(1:3).eq."ffd") then
             if (rec%w%all(iwa).ne.0) then
                call owffd( iwa, inode, +1,  dwr_z(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
             end if
          else if (extrap(1:4).eq."pspi") then
             if(present(sref)) then
                call owpspi(  iwa, inode, +1, dwr_z(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
             else
                call owpspi(  iwa, inode, +1, dwr_z(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
             end if
          end if
          ddwr(:,inode)=0
          if (renorm) then
             bwr(:,iz-1,inode)=bwr(:,iz-1,inode)+dwr_z(:,inode)
!             bwr(:,iz,inode)=bwr(:,iz,inode)+dwr_z(:,inode)
          end if
          if (rec%w%all(iwa).ne.0.) then
             call owscctrx(iwa,-1,ddwr(:,inode),bwr(:,iz-1,inode),vel(:,iz-1))
          end if
          GGG(:,iz,inode)=GGG(:,iz,inode)+conjg(ddwr(:,inode))*dwr_z(:,inode)
          
       end do

       if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
					" FROM NODE: ",inode
   end do
   !$OMP END PARALLEL DO

   call owfftdestroy()
   do i=1,nodes
      !$OMP parallel do private(ixw,iz)
      do iz=max(1,nzmin),nzmax
         do ixw=1,kxm%n
            deltaS(ixw,iz)=deltaS(ixw,iz)+GGG(ixw,iz,i)
         end do
      end do
      !$OMP END PARALLEL DO
   end do

   deallocate(dwr,dws)
else
   GG=0
   allocate(dwr(1,kxm%n,nodes));               dwr=0
   allocate(dws(1,kxm%n,nodes));               dws=0
   !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
    do iwa = 1,rec%w%n
       inode=omp_get_thread_num()+1; 
       ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=source(:,iwa)
       wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=recv(:,iwa)
       dws(1,:,inode)=0
       dwr(1,:,inode)=0
       bws(:,:,inode)=0.
       bwr(:,:,inode)=0.
       bws(:,1,inode)=ws(:,inode)
       bwr(:,1,inode)=wr(:,inode)
!!!!!!!!!!!!!!!!!!!!!!!
!
! Background wavefields
! bws and bwr
!
       do iz=max(2,nzmin),nzmax
          if (extrap(1:3).eq."ffd") then
             if (rec%w%all(iwa).ne.0) then
                call owffd( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
                call owffd( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
             end if
          else if (extrap(1:4).eq."pspi") then
             if (present(sref)) then
                call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
                call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
             else
                call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
                call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
             end if
          end if
          ws(:,inode)=ws(:,inode)*taper
          wr(:,inode)=wr(:,inode)*taper
          bws(:,iz,inode)=ws(:,inode)
          bwr(:,iz,inode)=wr(:,inode)
       end do
      do iz=max(2,nzmin),nzmax
          ddws(:,inode)=0
          if (renorm) then
             bws(:,iz-1,inode)=bws(:,iz-1,inode)+dws(1,:,inode)
!             bws(:,iz,inode)=bws(:,iz,inode)+dws(1,:,inode)
          end if
          if (rec%w%all(iwa).ne.0.) then
!             call owscctrx(iwa, 1,ddws(:,inode),bws(:,iz-1,inode),vel(:,iz-1))
             call owscctrx(iwa, 1,ddws(:,inode),bws(:,iz-1,inode),vel(:,iz-1))
          end if
          dws(1,:,inode)=dws(1,:,inode)+ddws(:,inode)*deltaS(:,iz-1)
          if (extrap(1:3).eq."ffd") then
             if (rec%w%all(iwa).ne.0) then
                call owffd( iwa, inode, 1,  dws(1,:,inode), vel(:,iz-1), tflds(:,inode), "sou")
             end if
          else if (extrap(1:4).eq."pspi") then
             if (present(sref)) then
                call owpspi(  iwa, inode, 1, dws(1,:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
             else
                call owpspi(  iwa, inode, 1, dws(1,:,inode), vel(:,iz-1), tflds(:,inode), "sou")
             end if
          end if
          dws(1,:,inode)=dws(1,:,inode)*taper
          if (exist_file("dkws")) zws(:,iz,inode)=dws(1,:,inode)
          call owimage(.false., iwa, dws(1,:,inode), bwr(:,iz,inode), GG(:,:,iz,inode),vel(:,iz))

          ddwr(:,inode)=0
          if (renorm) then
             bwr(:,iz-1,inode)=bwr(:,iz-1,inode)+dwr(1,:,inode)
!             bwr(:,iz,inode)=bwr(:,iz,inode)+dwr(1,:,inode)
          end if
          if (rec%w%all(iwa).ne.0.) then
             call owscctrx(iwa,-1,ddwr(:,inode),bwr(:,iz-1,inode),vel(:,iz-1))
!             call owscctrx(iwa,-1,ddwr(:,inode),bwr(:,iz,inode),vel(:,iz-1))
          end if
          dwr(1,:,inode)=dwr(1,:,inode)+ddwr(:,inode)*deltaS(:,iz-1)
          if (extrap(1:3).eq."ffd") then
             if (rec%w%all(iwa).ne.0) then
                call owffd( iwa, inode, -1,  dwr(1,:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
             end if
          else if (extrap(1:4).eq."pspi") then
             if (present(sref)) then
                call owpspi(  iwa, inode,-1, dwr(1,:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
             else
                call owpspi(  iwa, inode,-1, dwr(1,:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
             end if
          end if
          dwr(1,:,inode)=dwr(1,:,inode)*taper
          if (exist_file("dkwr")) zwr(:,iz,inode)=dwr(1,:,inode)
          call owimage(.false., iwa, bws(:,iz,inode), dwr(1,:,inode), GG(:,:,iz,inode),vel(:,iz-1))

          if (born2) then
             call owimage(.false., iwa, dws(1,:,inode), dwr(1,:,inode), GG(:,:,iz,inode),vel(:,iz-1))
          end if
       end do
       if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
					" FROM NODE: ",inode
       if (exist_file("bkwr")) then
          !$OMP ORDERED
          call ow_wvfld_io1(bwr(:,:,inode),"bkwr","rite")
          call ow_wvfld_io1(bws(:,:,inode),"bkws","rite")
          call ow_wvfld_io1(zwr(:,:,inode),"dkwr","rite")
          call ow_wvfld_io1(zws(:,:,inode),"dkws","rite")
         !$OMP END ORDERED       
       end if
   end do
   !$OMP END PARALLEL DO
   call owfftdestroy()
   do i=1,nodes
      !$OMP parallel do private(ixw,iz,ish)
      do iz=max(1,nzmin),nzmax
         do ish=1,nsecdim
            do ixw=1,kxm%n
               deltaI(ixw,ish,iz)=deltaI(ixw,ish,iz)+GG(ixw,ish,iz,i)
            end do
         end do
      end do
      !$OMP END PARALLEL DO
   end do
   deallocate(dwr,dws)
end if

6 format(A13,i3,A8,i3,A14,i3,A12,i3)
end subroutine
!!======================================================================
subroutine owwemva1(adj,taper,recv,source,vel,deltaS,deltaI,sref)

real, optional					:: sref(:,:,:)
logical						:: adj
complex						:: recv(:,:),source(:,:)
integer						:: iwa,iz,inode,i,ixw,ish,n0
real						:: taper(:),vel(kxm%n,img%z%n)
real						:: deltaS(kxm%n,img%z%n,sou%w%n)
real						:: deltaI(kxm%n,nsecdim,img%z%n)
real						:: GGG(kxm%n,img%z%n,sou%w%n)
real						:: GG(kxm%n,nsecdim,img%z%n,nodes)
complex						:: ws(kxm%n,nodes),wr(kxm%n,nodes)
complex						:: ddws(kxm%n,nodes),ddwr(kxm%n,nodes),temp(kxm%n,nodes)
complex						:: dws(kxm%n,img%z%n,nodes),dwr(kxm%n,img%z%n,nodes)
complex						:: bws(kxm%n,img%z%n,sou%w%n),bwr(kxm%n,img%z%n,sou%w%n)
complex						:: tflds(kxm%n,nodes),tfldr(kxm%n,nodes)
complex						:: dws_z(kxm%n,nodes),dwr_z(kxm%n,nodes)
complex						:: zddws(kxm%n,img%z%n,sou%w%n),zddwr(kxm%n,img%z%n,sou%w%n)
integer, external					:: omp_get_thread_num,omp_get_num_threads

!--------------------------------------------------
! Setting FFTW plans
tflds=0
call owfftws(tflds,nodes)
tfldr=0
call owfftwr(tfldr,nodes)

if (adj) then
   deltaS=0
  
   !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
   do iwa = 1,rec%w%n
       inode=omp_get_thread_num()+1
       ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=source(:,iwa)
       wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=recv(:,iwa)
       dws(:,:,inode)=0.
       dwr(:,:,inode)=0.
       bws(:,:,iwa)=0.
       bwr(:,:,iwa)=0.
       dws_z(:,inode)=0.
       dwr_z(:,inode)=0.
       ddwr(:,inode)=0
       ddws(:,inode)=0
       bws(:,1,iwa)=ws(:,inode)
       bwr(:,1,iwa)=wr(:,inode)

!!!!!!!!!!!!!!!!!!!!!!!
!
! Background wavefields
! bws and bwr
!
      do iz=max(2,nzmin),nzmax
         if (extrap(1:3).eq."ffd") then
            if (rec%w%all(iwa).ne.0) then
               call owffd( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
               call owffd( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
            end if
         else if (extrap(1:4).eq."pspi") then
            if (present(sref)) then
               call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
               call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
            else
               call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
               call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
            end if
         end if
          ws(:,inode)=ws(:,inode)*taper
          wr(:,inode)=wr(:,inode)*taper
          bws(:,iz,iwa)=ws(:,inode)
          bwr(:,iz,iwa)=wr(:,inode)
       end do
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Back project image perturbation
! 
! 
       do iz=img%z%n,max(1,nzmin),-1
          call owimage(.true.,  iwa, dws(:,iz,inode),  bwr(:,iz,iwa), deltaI(:,:,iz), vel(:,iz), .true. )
          call owimage(.true.,  iwa, bws(:,iz,iwa),  dwr(:,iz,inode), deltaI(:,:,iz), vel(:,iz), .false.)
          dws_z(:,inode)=dws_z(:,inode)+dws(:,iz,inode)
          dwr_z(:,inode)=dwr_z(:,inode)+dwr(:,iz,inode)
          dwr_z(:,inode)=dwr_z(:,inode)*taper
          dws_z(:,inode)=dws_z(:,inode)*taper

          if (extrap(1:3).eq."ffd") then
             if (rec%w%all(iwa)*10.ne.0) then
                call owffd( iwa, inode, -1,  dws_z(:,inode), vel(:,iz), tflds(:,inode), "sou")
             end if
          else if (extrap(1:4).eq."pspi") then
             if(present(sref)) then
                call owpspi(  iwa, inode, -1, dws_z(:,inode), vel(:,iz), tflds(:,inode), "sou", sref(:,:,iz))
             else
                call owpspi(  iwa, inode, -1, dws_z(:,inode), vel(:,iz), tflds(:,inode), "sou")
             end if
          end if
          ddws(:,inode)=0
          if (rec%w%all(iwa).ne.0.) then
             call owscctr (iwa,inode,1,ddws(:,inode),bws(:,iz,iwa),vel(:,iz),tflds(:,inode),"sou")
          end if

          if (extrap(1:3).eq."ffd") then
             if (rec%w%all(iwa).ne.0) then
                call owffd( iwa, inode, +1,  dwr_z(:,inode), vel(:,iz), tfldr(:,inode), "rec")
             end if
          else if (extrap(1:4).eq."pspi") then
             if(present(sref)) then
                call owpspi(  iwa, inode, +1, dwr_z(:,inode), vel(:,iz), tfldr(:,inode), "rec", sref(:,:,iz))
             else
                call owpspi(  iwa, inode, +1, dwr_z(:,inode), vel(:,iz), tfldr(:,inode), "rec")
             end if
          end if
          ddwr(:,inode)=0
          if (rec%w%all(iwa).ne.0.) then
             call owscctr (iwa,inode,-1,ddwr(:,inode),bwr(:,iz,iwa),vel(:,iz),tfldr(:,inode),"rec")
          end if

          deltaS(:,iz,iwa)=deltaS(:,iz,iwa)+conjg(ddws(:,inode))*dws_z(:,inode)
          deltaS(:,iz,iwa)=deltaS(:,iz,iwa)+conjg(ddwr(:,inode))*dwr_z(:,inode)
          zddwr(:,iz,iwa)=dwr_z(:,inode)
          zddws(:,iz,iwa)=dws_z(:,inode)
       end do

       if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
					" FROM NODE: ",inode
   end do
   !$OMP END PARALLEL DO
   
   if (renorm) then
      call ow_wvfld_io(bwr,  "bkwr","rite")
      call ow_wvfld_io(bws,  "bkws","rite")
      call ow_wvfld_io(zddwr,"dkwr","rite")
      call ow_wvfld_io(zddws,"dkws","rite")
   end if
   call owfftdestroy()
end if

6 format(A13,i3,A8,i3,A14,i3,A12,i3)
end subroutine
!!======================================================================
subroutine ow_wvfld_io_init(tag)
character*4    :: tag

   call to_history("n1",img%xm%n,tag); call to_history("d1",img%xm%d,tag); call to_history("o1",img%xm%o,tag)  
   call to_history("n2",nz,tag) ; call to_history("d2",img%z%d,tag) ; call to_history("o2",zmin,tag)
   call to_history("n3",1,tag); call to_history("d3",1,tag); call to_history("o3",0,tag)  
   call to_history("n4",1,tag); call to_history("d4",1,tag); call to_history("o4",1,tag)
   call to_history("n5",rec%w%n,tag) ; call to_history("d5",rec%w%d/(2*pi),tag) ; call to_history("o5",rec%w%o/(2*pi),tag)
   call to_history("n6",rec%xd%n,tag); call to_history("d6",1,tag); call to_history("o6",1,tag)
   call to_history("esize",8,tag)

end subroutine
!!======================================================================
subroutine ow_wvfld_io(CCC,tag,op)
character*4    :: tag,op
complex        :: CCC(:,:,:),DDD(img%xm%n,nz,sou%w%n)
integer        :: n0


if (op.eq."rite") then
   DDD=0
   n0 = (((idip-1)*rec%xd%d+rec%xd%o+rec%h%o-(padtraces+tapertr)*rec%h%d)-img%xm%o)/img%xm%d+1.5
   DDD(n0:kxm%n+n0-1,:,:)=CCC
   call srite(tag,DDD,8*img%xm%n*nz*rec%w%n)
else if (op.eq."reed") then
   call sreed(tag,CCC,8*img%xm%n*nz*rec%w%n)
end if

end subroutine
!!======================================================================
subroutine ow_wvfld_io1(CCC,tag,op)
character*4    :: tag,op
complex        :: CCC(:,:),DDD(img%xm%n,nz)
integer        :: n0


if (op.eq."rite") then
   DDD=0
   n0 = (((idip-1)*rec%xd%d+rec%xd%o+rec%h%o-(padtraces+tapertr)*rec%h%d)-img%xm%o)/img%xm%d+1.5
   DDD(n0:kxm%n+n0-1,:)=CCC
   call srite(tag,DDD,8*img%xm%n*nz)
else if (op.eq."reed") then
   call sreed(tag,CCC,8*img%xm%n*nz*rec%w%n)
end if

end subroutine
!======================================================================
subroutine owdso_renorm(adj,wgcig)!,ofcig)
logical						:: adj
real						:: wgcig(img%xm%n,nsecdim,img%z%n,sou%w%n)
integer						:: iz,ix,ish
real						:: h(nsecdim)

   do ish=1,nsecdim
      h(ish)=(ish-1)*dsecdim+osecdim
   end do
   h = (abs(h / maxval(h)))**expdso

   do iz=1,sou%w%n;do ish=1,nsecdim;do ix=1,img%xm%n
      wgcig(ix,ish,:,iz)=wgcig(ix,ish,:,iz)*h(ish)
   end do;end do;end do
end subroutine
!======================================================================
subroutine owwei_renorm(taper,recv,source,vel,image,sref)

real, optional					:: sref(:,:,:)
complex						:: recv(:,:),source(:,:)
integer						:: iwa,iz,inode,i,ixw,ish
real						:: taper(:),vel(kxm%n,img%z%n),a
real						:: image(:,:,:,:)
complex						:: ws(kxm%n,nodes),wr(kxm%n,nodes)
complex						:: tflds(kxm%n,nodes),tfldr(kxm%n,nodes)
integer, external					:: omp_get_thread_num,omp_get_num_threads

wr=0
ws=0


!--------------------------------------------------
! Setting FFTW plans
tflds=0
call owfftws(tflds,nodes)
tfldr=0
call owfftwr(tfldr,nodes)
!$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
   do iwa = 1,rec%w%n
      inode=omp_get_thread_num()+1
      ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=source(:,iwa)
      wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=recv(:,iwa)
      do iz=max(2,nzmin),nzmax
         if (extrap(1:3).eq."ffd") then
            if (rec%w%all(iwa).ne.0) then
               call owffd( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
               call owffd( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
            end if
         else if (extrap(1:4).eq."pspi") then
            if (present(sref)) then
               call owpspi_init(iz)
               call owpspi( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
               call owpspi( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
            else
               call owpspi_init(iz)
               call owpspi( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
               call owpspi( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
            end if
         end if
         ws(:,inode)=ws(:,inode)*taper
         wr(:,inode)=wr(:,inode)*taper
         call owimage(.false., iwa, ws(:,inode),  wr(:,inode), image(:,:,iz-nzmin+1,iwa), vel(:,iz-1))
      end do
      if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
				" FROM NODE: ",inode
   end do
!$OMP END PARALLEL DO
6 format(A13,i3,A8,i3,A14,i3,A12,i3)
call owfftdestroy()
end subroutine
!======================================================================
subroutine owwemva_renorm(adj,taper,recv,source,vel,deltaS,deltaI,sref)

real, optional					:: sref(:,:,:)
logical						:: adj
complex						:: recv(:,:),source(:,:)
integer						:: iwa,iz,inode,i,ixw,ish,n0
real						:: taper(:),vel(kxm%n,img%z%n)
real						:: deltaS(kxm%n,img%z%n)
real						:: deltaI(:,:,:,:)!,GG(kxm%n,nsecdim,img%z%n,nodes)
complex						:: ws(kxm%n,nodes),wr(kxm%n,nodes)
complex						:: ddws(kxm%n,nodes),ddwr(kxm%n,nodes)
complex 						:: dws(kxm%n,nodes),dwr(kxm%n,nodes)
complex 						:: zws(kxm%n,img%z%n,nodes),zwr(kxm%n,img%z%n,nodes)
complex						:: bws(kxm%n,img%z%n,nodes),bwr(kxm%n,img%z%n,nodes)
complex						:: tflds(kxm%n,nodes),tfldr(kxm%n,nodes)
integer, external					:: omp_get_thread_num,omp_get_num_threads

deltaI=0;!GG=0
!--------------------------------------------------
! Setting FFTW plans
tflds=0
call owfftws(tflds,nodes)
tfldr=0
call owfftwr(tfldr,nodes)

   !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
    do iwa = 1,rec%w%n
       inode=omp_get_thread_num()+1; 
       ws(:,inode)=0;wr(:,inode)=0.
       ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=source(:,iwa)
       wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=recv(:,iwa)
       dws(:,inode)=0
       dwr(:,inode)=0
       zws(:,:,inode)=0
       zwr(:,:,inode)=0
       bws(:,:,inode)=0.
       bwr(:,:,inode)=0.
       bws(:,1,inode)=ws(:,inode)
       bwr(:,1,inode)=wr(:,inode)
!!!!!!!!!!!!!!!!!!!!!!!
!
! Background wavefields
! bws and bwr
!
       do iz=max(2,nzmin),nzmax
          if (extrap(1:3).eq."ffd") then
             if (rec%w%all(iwa).ne.0) then
                call owffd( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
                call owffd( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
             end if
          else if (extrap(1:4).eq."pspi") then
             if (present(sref)) then
                call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
                call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
             else
                call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
                call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
             end if
          end if
          ws(:,inode)=ws(:,inode)*taper
          wr(:,inode)=wr(:,inode)*taper
          bws(:,iz,inode)=ws(:,inode)
          bwr(:,iz,inode)=wr(:,inode)
       end do 
       do iz=max(2,nzmin),nzmax
          ddws(:,inode)=0
          bws(:,iz-1,inode)=bws(:,iz-1,inode)+dws(:,inode)
          if (rec%w%all(iwa).ne.0.) then
             call owscctrx(iwa,+1,ddws(:,inode),bws(:,iz-1,inode),vel(:,iz-1))
!              call owscctr (iwa,inode,1,ddws(:,inode),bws(:,iz-1,inode),vel(:,iz-1),tflds(:,inode),"sou")
          end if
          dws(:,inode)=dws(:,inode)+ddws(:,inode)*deltaS(:,iz-1)
          if (extrap(1:3).eq."ffd") then
             if (rec%w%all(iwa).ne.0) then
                call owffd( iwa, inode, 1,  dws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
             end if
          else if (extrap(1:4).eq."pspi") then
             if (present(sref)) then
                call owpspi(  iwa, inode, 1, dws(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
             else
                call owpspi(  iwa, inode, 1, dws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
             end if
          end if
          dws(:,inode)=dws(:,inode)*taper
          if (exist_file("dkws")) zws(:,iz,inode)=dws(:,inode)
          call owimage(.false., iwa, dws(:,inode), bwr(:,iz,inode), deltaI(:,:,iz,iwa),vel(:,iz))

          ddwr(:,inode)=0
          bwr(:,iz-1,inode)=bwr(:,iz-1,inode)+dwr(:,inode)
          if (rec%w%all(iwa).ne.0.) then
             call owscctrx(iwa,-1,ddwr(:,inode),bwr(:,iz-1,inode),vel(:,iz-1))
!              call owscctr (iwa,inode,-1,ddwr(:,inode),bwr(:,iz-1,inode),vel(:,iz-1),tfldr(:,inode),"rec")
          end if
          dwr(:,inode)=dwr(:,inode)+ddwr(:,inode)*deltaS(:,iz-1)
          if (extrap(1:3).eq."ffd") then
             if (rec%w%all(iwa).ne.0) then
                call owffd( iwa, inode, -1,  dwr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
             end if
          else if (extrap(1:4).eq."pspi") then
             if (present(sref)) then
                call owpspi(  iwa, inode,-1, dwr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
             else
                call owpspi(  iwa, inode,-1, dwr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
             end if
          end if
          dwr(:,inode)=dwr(:,inode)*taper
          if (exist_file("dkwr")) zwr(:,iz,inode)=dwr(:,inode)
          call owimage(.false., iwa, bws(:,iz,inode), dwr(:,inode), deltaI(:,:,iz,iwa),vel(:,iz-1))

       end do
       if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
					" FROM NODE: ",inode

       if (exist_file("bkwr")) then
          !$OMP ORDERED
          call ow_wvfld_io1(bwr(:,:,inode),"bkwr","rite")
          call ow_wvfld_io1(bws(:,:,inode),"bkws","rite")
          call ow_wvfld_io1(zwr(:,:,inode),"dkwr","rite")
          call ow_wvfld_io1(zws(:,:,inode),"dkws","rite")
         !$OMP END ORDERED       
       end if

   end do
   !$OMP END PARALLEL DO
   call owfftdestroy()

6 format(A13,i3,A8,i3,A14,i3,A12,i3)
end subroutine
!=========================================================================
!subroutine owwemva_adj(taper,recv,source,vel,f,deltaS,deltaI,sref)
!
!
!real, optional					:: sref(:,:,:)
!double precision                                            :: f
!complex						:: recv(:,:),source(:,:)
!integer						:: iwa,iz,inode,i,ixw,ish
!real						:: taper(:),vel(kxm%n,img%z%n),fun(nodes)
!real						:: deltaS(kxm%n,img%z%n)
!real						:: deltaI(kxm%n,nsecdim,img%z%n)
!real						:: GGG(kxm%n,img%z%n,nodes)
!real						:: GG(kxm%n,nsecdim,img%z%n,nodes)
!complex						:: ws(kxm%n,nodes),wr(kxm%n,nodes)
!complex						:: ddws(kxm%n,nodes),ddwr(kxm%n,nodes)
!complex		 				:: dws(kxm%n,img%z%n,nodes),dwr(kxm%n,img%z%n,nodes)
!complex						:: bws(kxm%n,img%z%n,nodes),bwr(kxm%n,img%z%n,nodes)
!complex						:: tflds(kxm%n,nodes),tfldr(kxm%n,nodes)
!complex						:: dws_z(kxm%n,nodes),dwr_z(kxm%n,nodes)
!complex						:: zddws(kxm%n,img%z%n,sou%w%n),zddwr(kxm%n,img%z%n,sou%w%n)
!complex						:: ddwsz(kxm%n,img%z%n,sou%w%n),ddwrz(kxm%n,img%z%n,sou%w%n)
!integer, external					:: omp_get_thread_num,omp_get_num_threads
!
!wr=0
!ws=0
!ddwr=0
!ddws=0
!bwr=0
!bws=0
!
!!--------------------------------------------------
!! Setting FFTW plans
!tflds=0
!call owfftws(tflds,nodes)
!tfldr=0
!call owfftwr(tfldr,nodes)
!
!   GGG=0
!   dwr_z=0
!   dws_z=0
!   dwr=0
!   dws=0
!  
!   !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
!   do iwa = 1,rec%w%n
!       inode=omp_get_thread_num()+1
!       ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=source(:,iwa)
!       wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=recv(:,iwa)
!       dws(:,:,inode)=0.
!       dwr(:,:,inode)=0.
!       bws(:,:,inode)=0.
!       bwr(:,:,inode)=0.
!       dws_z(:,inode)=0.
!       dwr_z(:,inode)=0.
!       ddwr(:,inode)=0
!       ddws(:,inode)=0
!       bws(:,1,inode)=ws(:,inode)
!       bwr(:,1,inode)=wr(:,inode)
!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Background wavefields
!! bws and bwr
!!
!
!      do iz=max(2,nzmin),nzmax
!         if (extrap(1:3).eq."ffd") then
!            if (rec%w%all(iwa).ne.0) then
!               call owffd( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
!               call owffd( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
!            end if
!         else if (extrap(1:4).eq."pspi") then
!            if (present(sref)) then
!               call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
!               call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
!            else
!               call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
!               call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
!            end if
!         end if
!          ws(:,inode)=ws(:,inode)*taper
!          wr(:,inode)=wr(:,inode)*taper
!          bws(:,iz,inode)=ws(:,inode)
!          bwr(:,iz,inode)=wr(:,inode)
!          call owimage(.false., iwa, ws(:,inode),  wr(:,inode), GG(:,:,iz-nzmin+1,inode), vel(:,iz-1))
!       end do
!
!       call owdso(.false. , GG(:,:,:,inode))
!       call owdso(.true.  , GG(:,:,:,inode))
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Computing the objective function
!!
!       fun=dble(0.)
!       do iz=1,img%z%n
!          do ish=1,img%sh%n
!             do ixw=1,img%xm%n
!                fun(inode) = fun(inode) + GG(ixw,ish,iz-nzmin+1,inode)*GG(ixw,ish,iz-nzmin+1,inode)
!             end do
!          end do
!       end do
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Back project slowness perturbation
!! 
!! 
!       do iz=img%z%n,max(1,nzmin),-1
!          call owimage(.true.,  iwa, dws(:,iz,inode),  bwr(:,iz,inode), GG(:,:,iz,inode), vel(:,iz), .true. )
!          dws_z(:,inode)=dws_z(:,inode)+dws(:,iz,inode)
!          dws_z(:,inode)=dws_z(:,inode)*taper
!          if (extrap(1:3).eq."ffd") then
!             if (rec%w%all(iwa).ne.0) then
!                call owffd( iwa, inode, -1,  dws_z(:,inode), vel(:,iz), tflds(:,inode), "sou")
!             end if
!          else if (extrap(1:4).eq."pspi") then
!             if(present(sref)) then
!                call owpspi(  iwa, inode, -1, dws_z(:,inode), vel(:,iz), tflds(:,inode), "sou", sref(:,:,iz))
!             else
!                call owpspi(  iwa, inode, -1, dws_z(:,inode), vel(:,iz), tflds(:,inode), "sou")
!             end if
!          end if
!          ddws(:,inode)=0
!          if (rec%w%all(iwa).ne.0.) then
!!             call owscctr(iwa, 1,ddws(:,inode),bws(:,iz,inode),vel(:,iz))
!              call owscctr (iwa,inode,1,ddws(:,inode),bws(:,iz,inode),vel(:,iz),tflds(:,inode),"sou")
!          end if
!          if (exist_file("ddws")) then
!             zddws(:,iz,iwa)=dws_z(:,inode)
!             ddwsz(:,iz,iwa)=ddws(:,inode)
!          end if
!          GGG(:,iz,inode)=GGG(:,iz,inode)+conjg(ddws(:,inode))*dws_z(:,inode)!/rec%w%n
!
!          call owimage(.true.,  iwa, bws(:,iz,inode),  dwr(:,iz,inode), GG(:,:,iz,inode), vel(:,iz), .false.)
!          dwr_z(:,inode)=dwr_z(:,inode)+dwr(:,iz,inode)
!          dwr_z(:,inode)=dwr_z(:,inode)*taper
!          if (extrap(1:3).eq."ffd") then
!             if (rec%w%all(iwa).ne.0) then
!                call owffd( iwa, inode, +1,  dwr_z(:,inode), vel(:,iz), tfldr(:,inode), "rec")
!             end if
!          else if (extrap(1:4).eq."pspi") then
!             if(present(sref)) then
!                call owpspi(  iwa, inode, +1, dwr_z(:,inode), vel(:,iz), tfldr(:,inode), "rec", sref(:,:,iz))
!             else
!                call owpspi(  iwa, inode, +1, dwr_z(:,inode), vel(:,iz), tfldr(:,inode), "rec")
!             end if
!          end if
!          ddwr(:,inode)=0
!          if (rec%w%all(iwa).ne.0.) then
!!             call owscctr(iwa,-1,ddwr(:,inode),bwr(:,iz,inode),vel(:,iz))
!              call owscctr (iwa,inode,-1,ddwr(:,inode),bwr(:,iz,inode),vel(:,iz),tflds(:,inode),"sou")
!          end if
!          if (exist_file("ddwr")) then
!             zddwr(:,iz,iwa)=dwr_z(:,inode)
!             ddwrz(:,iz,iwa)=ddwr(:,inode)
!          end if
!          GGG(:,iz,inode)=GGG(:,iz,inode)+conjg(ddwr(:,inode))*dwr_z(:,inode)!/rec%w%n
!       end do
!
!       if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
!					" FROM NODE: ",inode
!   end do
!   !$OMP END PARALLEL DO
!   call owfftdestroy()
!   do i=1,nodes
!      f = f + dble(fun(i))
!      !$OMP parallel do private(ixw,iz)
!      do iz=max(1,nzmin),nzmax
!         do ixw=1,kxm%n
!            deltaS(ixw,iz)=deltaS(ixw,iz)+GGG(ixw,iz,i)
!            do ish=1,nsecdim
!               deltaI(ixw,ish,iz)=deltaI(ixw,ish,iz)+GG(ixw,ish,iz,i)
!            end do
!         end do
!      end do
!      !$OMP END PARALLEL DO
!   end do
!6 format(A13,i3,A8,i3,A14,i3,A12,i3)
!end subroutine 
!=============================================================
end module
!
!if (adj) then
!   deltaS=0
!   allocate(dws(kxm%n,img%z%n,nodes),dwr(kxm%n,img%z%n,nodes))  
!   !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
!   do iwa = 1,rec%w%n
!       inode=omp_get_thread_num()+1
!       ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=source(:,iwa)
!       wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode)=recv(:,iwa)
!       dws(:,:,inode)=0.
!       dwr(:,:,inode)=0.
!       bws(:,:,iwa)=0.
!       bwr(:,:,iwa)=0.
!       dws_z(:,inode)=0.
!       dwr_z(:,inode)=0.
!       ddwr(:,inode)=0
!       ddws(:,inode)=0
!       bws(:,1,iwa)=ws(:,inode)
!       bwr(:,1,iwa)=wr(:,inode)
!
!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Background wavefields
!! bws and bwr
!!
!      do iz=max(2,nzmin),nzmax
!         if (extrap(1:3).eq."ffd") then
!            if (rec%w%all(iwa).ne.0) then
!               call owffd( iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
!               call owffd( iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
!            end if
!         else if (extrap(1:4).eq."pspi") then
!            if (present(sref)) then
!               call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
!               call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
!            else
!               call owpspi(  iwa, inode, 1,  ws(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
!               call owpspi(  iwa, inode,-1,  wr(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
!            end if
!         end if
!          ws(:,inode)=ws(:,inode)*taper
!          wr(:,inode)=wr(:,inode)*taper
!          bws(:,iz,iwa)=ws(:,inode)
!          bwr(:,iz,iwa)=wr(:,inode)
!       end do
!
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Back project image perturbation
!! 
!! 
!       do iz=img%z%n,max(2,nzmin),-1
!
!          call owimage(.true.,  iwa, dws(:,iz,inode),  bwr(:,iz,iwa), deltaI(:,:,iz,iwa), vel(:,iz-1), .true. )
!          dws_z(:,inode)=dws_z(:,inode)+dws(:,iz,inode)
!          dws_z(:,inode)=dws_z(:,inode)*taper
!          if (extrap(1:3).eq."ffd") then
!             if (rec%w%all(iwa)*10.ne.0) then
!                call owffd( iwa, inode, -1,  dws_z(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
!             end if
!          else if (extrap(1:4).eq."pspi") then
!             if(present(sref)) then
!                call owpspi(  iwa, inode, -1, dws_z(:,inode), vel(:,iz-1), tflds(:,inode), "sou", sref(:,:,iz-1))
!             else
!                call owpspi(  iwa, inode, -1, dws_z(:,inode), vel(:,iz-1), tflds(:,inode), "sou")
!             end if
!          end if
!          ddws(:,inode)=0
!          if (rec%w%all(iwa).ne.0.) then
!             call owscctr (iwa,inode,1,ddws(:,inode),bws(:,iz,iwa),vel(:,iz-1),tflds(:,inode),"sou")
!          end if
!!          deltaS(:,iz,iwa)=deltaS(:,iz,iwa)+conjg(ddws(:,inode))*dws_z(:,inode)!/rec%w%n
!          deltaS(:,iz)=deltaS(:,iz)+conjg(ddws(:,inode))*dws_z(:,inode)!/rec%w%n
!
!          call owimage(.true.,  iwa, bws(:,iz,iwa),  dwr(:,iz,inode), deltaI(:,:,iz,iwa), vel(:,iz-1), .false.)
!          dwr_z(:,inode)=dwr_z(:,inode)+dwr(:,iz,inode)
!          dwr_z(:,inode)=dwr_z(:,inode)*taper
!          if (extrap(1:3).eq."ffd") then
!             if (rec%w%all(iwa).ne.0) then
!                call owffd( iwa, inode, +1,  dwr_z(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
!             end if
!          else if (extrap(1:4).eq."pspi") then
!             if(present(sref)) then
!                call owpspi(  iwa, inode, +1, dwr_z(:,inode), vel(:,iz-1), tfldr(:,inode), "rec", sref(:,:,iz-1))
!             else
!                call owpspi(  iwa, inode, +1, dwr_z(:,inode), vel(:,iz-1), tfldr(:,inode), "rec")
!             end if
!          end if
!          ddwr(:,inode)=0
!          if (rec%w%all(iwa).ne.0.) then
!             call owscctr (iwa,inode,-1,ddwr(:,inode),bwr(:,iz,iwa),vel(:,iz-1),tfldr(:,inode),"rec")
!          end if
!!          deltaS(:,iz,iwa)=deltaS(:,iz,iwa)+conjg(ddwr(:,inode))*dwr_z(:,inode)!/rec%w%n
!          deltaS(:,iz)=deltaS(:,iz)+conjg(ddwr(:,inode))*dwr_z(:,inode)!/rec%w%n
!          
!       end do
!       if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
!					" FROM NODE: ",inode
!   end do
!   !$OMP END PARALLEL DO
!   call owfftdestroy()
!   deallocate(dwr,dws)
!   call ow_wvfld_io(bwr,  "bkwr","rite")
!   call ow_wvfld_io(bws,  "bkws","rite")
!else
